﻿#include <cigui/core/App.hpp>

#include <iostream>
#include <LLGL/Utils/VertexFormat.h>

namespace cig
{
	App::App()
	{
		// Rectangle
		m_Vertices.emplace_back(Vertex{ { 0.5f, -0.5f }, {    0, 255,   0, 255 } });
		m_Vertices.emplace_back(Vertex{ { -0.5f, -0.5f }, { 255,   0,   0, 255 } });
		m_Vertices.emplace_back(Vertex{ { -0.5f,  0.5f }, {   0,   0, 255, 255 } });
		m_Vertices.emplace_back(Vertex{ { 0.5f,  0.5f }, {  255,   0,   0, 255 } });
		m_Vertices.emplace_back(Vertex{ { -0.5f, -0.5f }, {   0, 255,   0, 255 } });
		m_Vertices.emplace_back(Vertex{ { -0.5f,  0.5f }, {   0,   0, 255, 255 } });

		for (const char* module : { "Direct3D12", "Direct3D11", "Vulkan", "OpenGL" }) {
			LLGL::Report report;
			m_RenderSystem = LLGL::RenderSystem::Load(module, &report);
			if (m_RenderSystem == nullptr) {
				LLGL::Log::Errorf("%s", report.GetText());
			}
		}
		if (m_RenderSystem == nullptr) {
			std::cout << "No renderer found!" << std::endl;
			return;
		}

		LLGL::SwapChainDescriptor swapChainDesc;
		swapChainDesc.resolution  = { 800, 600 }; // Framebuffer resolution of 800x600 pixels
		swapChainDesc.fullscreen  = false;						// No fullscreen, use windowed mode
		swapChainDesc.samples     = 8;							// 8 samples for anti-aliasing
		m_SwapChain = m_RenderSystem->CreateSwapChain(swapChainDesc);

		LLGL::VertexFormat vertexFormat;
		vertexFormat.AppendAttribute({ "position", LLGL::Format::RG32Float  });
		vertexFormat.AppendAttribute({ "color",    LLGL::Format::RGBA8UNorm });

		LLGL::BufferDescriptor VBufferDesc;
		VBufferDesc.size          = sizeof(Vertex) * m_Vertices.size();	// Size (in bytes) of the buffer
		VBufferDesc.bindFlags     = LLGL::BindFlags::VertexBuffer;		// Use for vertex buffer binding
		VBufferDesc.vertexAttribs = vertexFormat.attributes;			// Vertex buffer attributes
		m_VertexBuffer = m_RenderSystem->CreateBuffer(VBufferDesc, m_Vertices.data());

		LLGL::ShaderDescriptor VSDesc, FSDesc;
		if (ShaderLanguageIsSupported(LLGL::ShadingLanguage::HLSL)) {
			VSDesc = { LLGL::ShaderType::Vertex,   "shader/MyShader.hlsl", "VMain", "vs_4_0" };
			FSDesc = { LLGL::ShaderType::Fragment, "shader/MyShader.hlsl", "PMain", "ps_4_0" };
		} else if (ShaderLanguageIsSupported(LLGL::ShadingLanguage::GLSL)) {
			VSDesc = { LLGL::ShaderType::Vertex,   "shader/MyShader.vert" };
			FSDesc = { LLGL::ShaderType::Fragment, "shader/MyShader.frag" };
		} else {
			std::cout << "Shading language not supported!" << std::endl;
			return;
		}
		VSDesc.vertex.inputAttribs = vertexFormat.attributes;

		m_ShaderPack.first  = m_RenderSystem->CreateShader(VSDesc);
		m_ShaderPack.second = m_RenderSystem->CreateShader(FSDesc);

		for (LLGL::Shader* shader : { m_ShaderPack.first, m_ShaderPack.second }) {
			if (const LLGL::Report* report = shader->GetReport()) {
				if (report->HasErrors()) {
					LLGL::Log::Errorf("Shader compile errors:\n%s", report->GetText());
				} else {
					LLGL::Log::Printf("Shader compile warnings:\n%s", report->GetText());
				}
			}
		}

		LLGL::GraphicsPipelineDescriptor pipelineDesc;
		pipelineDesc.vertexShader                  = m_ShaderPack.first;
		pipelineDesc.fragmentShader                = m_ShaderPack.second;
		pipelineDesc.rasterizer.multiSampleEnabled = swapChainDesc.samples > 1;

		m_Pipeline = m_RenderSystem->CreatePipelineState(pipelineDesc);

		m_CmdBuffer = m_RenderSystem->CreateCommandBuffer(LLGL::CommandBufferFlags::ImmediateSubmit);
	}

	App::~App() {}


	bool App::ShaderLanguageIsSupported(LLGL::ShadingLanguage lang)
	{
		const auto& supportedLangs = m_RenderSystem->GetRenderingCaps().shadingLanguages;
		return std::find(supportedLangs.begin(), supportedLangs.end(), lang) != supportedLangs.end();
	}

	App& App::GetInstance()
	{
		static App instance;
		return instance;
	}

	void App::Run()
	{
		if (!m_RenderSystem)
			return;
		m_Window = &LLGL::CastTo<LLGL::Window>(m_SwapChain->GetSurface());
		m_Window->SetTitle("Hello Triangle");
		m_Window->Show();

		while (!m_Window->HasQuit()) {
			m_Window->ProcessEvents();

			m_CmdBuffer->Begin();
			m_CmdBuffer->SetVertexBuffer(*m_VertexBuffer);

			m_CmdBuffer->BeginRenderPass(*m_SwapChain);

			m_CmdBuffer->SetViewport(m_SwapChain->GetResolution());

			m_CmdBuffer->Clear(LLGL::ClearFlags::Color);

			m_CmdBuffer->SetPipelineState(*m_Pipeline);

			m_CmdBuffer->Draw(m_Vertices.size(), 0);

			m_CmdBuffer->EndRenderPass();

			m_CmdBuffer->End();


			m_SwapChain->Present();
		}
	}
}
